from django import forms
from django.forms import BaseForm
from django.forms.fields import Field
from django.forms.widgets import MediaDefiningClass
from django.utils.translation import gettext_lazy as _

from .utils import labelize


class DeclarativeFieldsMetaclass(MediaDefiningClass):
    """Collect Fields declared on the base classes."""

    def __new__(mcs, name, bases, attrs):
        # Collect fields from current class.
        current_fields = []
        for key, value in list(attrs.items()):
            if isinstance(value, Field):
                current_fields.append((key, value))
                attrs.pop(key)

        meta = attrs.get('Meta', None)
        dynamic_fields = getattr(meta, 'dynamic_fields', '')
        if meta and dynamic_fields:
            if isinstance(dynamic_fields, dict):
                for field_name, field_label in dynamic_fields.items():
                    if any([field_name in key for key in current_fields]):
                        continue
                    current_fields.append((field_name, forms.CharField(label=_(field_label))))
            elif isinstance(dynamic_fields, list) or isinstance(dynamic_fields, tuple):
                for field_name in dynamic_fields:
                    if any([field_name in key for key in current_fields]):
                        continue
                    current_fields.append((field_name, forms.CharField(label=labelize(field_name))))
            else:
                dynamic_fields = dict(current_fields).keys()

            for field_name, field in current_fields:
                assert field_name in dynamic_fields, (
                    "The field '{field_name}' was declared on DynamicForm "
                    "{form_class}, but has not been included in the "
                    "'dynamic_fields' option.".format(
                        field_name=field_name,
                        form_class=name
                    )
                )

            current_fields = sorted(current_fields, key=lambda pair: list(dynamic_fields).index(pair[0]))

        attrs['declared_fields'] = dict(current_fields)
        new_class = super(DeclarativeFieldsMetaclass, mcs).__new__(mcs, name, bases, attrs)

        # Walk through the MRO.
        declared_fields = {}
        for base in reversed(new_class.__mro__):
            # Collect fields from base class.
            if hasattr(base, 'declared_fields'):
                declared_fields.update(base.declared_fields)

            # Field shadowing.
            for attr, value in base.__dict__.items():
                if value is None and attr in declared_fields:
                    declared_fields.pop(attr)

        new_class.base_fields = declared_fields
        new_class.declared_fields = declared_fields

        return new_class


class DynamicForm(BaseForm, metaclass=DeclarativeFieldsMetaclass):
    "A collection of Fields, plus their associated data."
    # This is a separate class from BaseForm in order to abstract the way
    # self.fields is specified. This class (Form) is the one that does the
    # fancy metaclass stuff purely for the semantic sugar -- it allows one
    # to define a form using declarative syntax.
    # BaseForm itself has no way of designating self.fields.



from django.core.paginator import InvalidPage
from django.http import Http404, HttpResponseRedirect, JsonResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView, ListView
from django.views.generic.edit import DeleteView, FormView

from . import status
from .mixins import BaseViewMixin
from .paginators import ClientAPIPagination
from .settings import api_client_settings
from .utils import clean_url


class ClientMethodMixin:
    client_method = None
    client_initial_method = None

    def get_client_method(self):
        return self.client_method

    def get_client_initial_method(self):
        return self.client_initial_method


class ClientAPIFormView(FormView):
    slug_field = api_client_settings.configs['SLUG_FIELD']
    slug_url_kwarg = api_client_settings.configs['SLUG_FIELD']

    def send_cleaned_data(self, response, form):
        if response.raw.status_code not in [status.HTTP_201_CREATED, status.HTTP_200_OK]:
            errors = response.raw.json()
            if not isinstance(errors, dict):
                errors = {'non_field_errors': errors}
            for error_key in errors:
                if error_key == 'non_field_errors':
                    form.add_error(None, errors[error_key])
                    continue
                form.add_error(error_key, errors[error_key])

            if self.request.is_ajax():
                return JsonResponse({'errors': form.errors})
            return self.form_invalid(form)
        if self.request.is_ajax():
            return JsonResponse(response.as_dict())
        return super().form_valid(form)

    def form_invalid(self, form):
        if self.request.is_ajax():
            if form.errors:
                return JsonResponse({'status': False, 'errors': form.errors})
        return super().form_invalid(form)


class ClientAPIAuthenticatedCreateView(ClientMethodMixin, ClientAPIFormView):

    def form_valid(self, form):
        client_method = self.get_client_method()
        response = client_method(data=form.cleaned_data)
        return self.send_cleaned_data(response, form)


class ClientAPIAuthenticatedUpdateView(ClientMethodMixin, ClientAPIFormView):
    partial = False

    def get_api_object_response(self):
        client_initial_method = self.get_client_initial_method()
        if not client_initial_method:
            return None
        return client_initial_method(self.kwargs.get(self.slug_field))

    def get_initial(self):
        api_object = self.get_api_object_response()
        if not api_object:
            return None
        return api_object.as_dict()

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        api_object = self.get_api_object_response()
        if api_object:
            context['object'] = api_object.as_obj()
        return context

    def form_valid(self, form):
        params = {
            'data': form.cleaned_data,
            'partial': self.partial
        }
        client_method = self.get_client_method()
        response = client_method(self.kwargs.get(self.slug_field), **params)
        return self.send_cleaned_data(response, form)


class ClientAPIAuthenticatedListView(ClientMethodMixin, ListView):
    http_method_names = ['get']
    paginate_by = api_client_settings.configs['PAGE_SIZE']
    paginator_class = ClientAPIPagination
    client_method_result_key = 'results'
    api_filters = []
    page_base_url = ''
    page_title = ''

    def get_paginate_by(self):
        """
        Get the number of items to paginate by, or ``None`` for no pagination.
        """
        return int(self.request.GET.get('paginate_by', self.paginate_by))

    def get_page_base_url(self):
        return self.page_base_url or self.request.path

    def _paginate_queryset(self, client_method, page_size, **kwargs):
        """Paginate the queryset, if needed."""
        extra_params = kwargs.get('extra_params', {})
        paginator = self.paginator_class(client_method, page_size, extra_params=extra_params)
        page_kwarg = self.page_kwarg
        page = self.kwargs.get(page_kwarg) or self.request.GET.get(page_kwarg) or 1
        try:
            page_number = int(page)
        except ValueError:
            if page == 'last':
                page_number = paginator.num_pages
            else:
                raise Http404(_('Page is not “last”, nor can it be converted to an int.'))
        try:
            page = paginator.page(page_number)
            return paginator, page, page.object_list, page.has_other_pages()
        except InvalidPage as e:
            raise Http404(_('Invalid page (%(page_number)s): %(message)s') % {
                'page_number': page_number,
                'message': str(e)
            })

    def get_api_params(self, request):
        default_api_filters = ['search']
        default_api_filters.extend(self.api_filters)
        api_filters = {}
        for param in default_api_filters:
            param_value = request.GET.get(param, '')
            if param_value:
                api_filters[param] = param_value
        return api_filters

    def get_context_data(self, **kwargs):
        filter_params = kwargs.pop('filter_params', False)
        paginate_by = self.get_paginate_by()
        if paginate_by:
            paginator, page, queryset, is_paginated = self._paginate_queryset(self.get_client_method(),
                                                                              paginate_by,
                                                                              extra_params=kwargs)
            context = {
                'paginator': paginator,
                'page_obj': page,
                'is_paginated': is_paginated,
                'object_list': queryset
            }
        else:
            context = {
                'paginator': None,
                'page_obj': None,
                'is_paginated': False,
                'object_list': self.object_list
            }

        search_url = clean_url(self.request.get_full_path())
        context.update({
            'search': self.request.GET.get('search', ''),
            'search_url': search_url,
            'filter_params': filter_params,
            'page_base_url': self.get_page_base_url(),
            'range_pagination': [x for x in range(20, 220, 20)],
            'paginate_by': paginate_by,
            'page_title': self.page_title
        })

        pagination_params = ['search', 'paginate_by', 'ordering'] + self.api_filters
        context['append_param'] = '&' if any(
            [True for x in pagination_params if x in self.request.GET.keys()]) else '?'

        return context

    def get(self, request, *args, **kwargs):
        params = self.get_api_params(request)
        if params:
            params['filter_params'] = True

        context = self.get_context_data(**params)
        self.object_list = context['object_list']

        if params:
            context.update(params)

        return self.render_to_response(context)


class ClientAPIAuthenticatedDetailView(ClientMethodMixin, DetailView):
    slug_field = api_client_settings.configs['SLUG_FIELD']
    slug_url_kwarg = api_client_settings.configs['SLUG_FIELD']

    def get(self, request, *args, **kwargs):
        client_method = self.get_client_method()
        response = client_method(self.kwargs.get(self.slug_field))
        self.object = response.as_obj()
        context = self.get_context_data()
        return self.render_to_response(context)


class ClientAPIAuthenticatedDeleteView(ClientMethodMixin, DeleteView):
    slug_field = api_client_settings.configs['SLUG_FIELD']
    slug_url_kwarg = api_client_settings.configs['SLUG_FIELD']

    def post(self, request, *args, **kwargs):
        client_method = self.get_client_method()
        client_method(self.kwargs.get(self.slug_field))
        success_url = self.get_success_url()
        return HttpResponseRedirect(success_url)


class ClientAPICreateView(BaseViewMixin, ClientAPIAuthenticatedCreateView):
    pass


class ClientAPIDetailView(BaseViewMixin, ClientAPIAuthenticatedDetailView):
    pass


class ClientAPIUpdateView(BaseViewMixin, ClientAPIAuthenticatedUpdateView):
    pass


class ClientAPIListView(BaseViewMixin, ClientAPIAuthenticatedListView):
    pass


class ClientAPIDeleteView(BaseViewMixin, ClientAPIAuthenticatedDeleteView):
    pass



from django.contrib.auth.mixins import LoginRequiredMixin
from django.views.generic.base import ContextMixin


class PageTitleMixin(ContextMixin):
    page_title = ''

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['page_title'] = self.page_title
        return context


class BaseViewMixin(PageTitleMixin, LoginRequiredMixin):
    pass



import json
from urllib.parse import urljoin

import requests
from django.utils.translation import gettext_lazy as _
from requests.exceptions import ConnectionError as RequestsConnectionError, ReadTimeout, Timeout

from django_api_client.utils import json_converter
from .exceptions import APIEndpointMissingArgument, APIUnreachableOrOffline
from .factories import ResponseFactory
from .validators import validate_status_code


class RequestsTransport:
    def __init__(self, headers):
        self.session = requests.Session()
        self.session.headers.update(headers)

    def make_request(self, method_name, endpoint, **kwargs):
        method = getattr(self.session, method_name.lower())
        return method(endpoint, **kwargs)


class BaseAPI:
    """
    BaseAPI Abstract object.
    Attributes:
        base_url: URL from API
        timeout: Request timeout
        token:  API Access token
        transport: Transport class to able to use Requests
    """

    def __init__(
            self,
            api_name,
            base_url,
            access_token,
            access_token_type='Bearer',
            authentication_method='header',
            authentication_url_extra_params=None,
            authentication_url_key=None,
            locale='',
            timeout=3,
            transport_class=None,
            **kwargs):
        """
        Args:
            api_name (str): API name
            base_url (str): Provides the environment for  API.
            access_token (str): Specifies the access token.
            access_token_type (str): Type of the token, could be Bearer, Token, etc. Default is Bearer
            authentication_method (str): Method used to authenticate. Default is 'header'
            authentication_url_extra_params (dict): Extra params to authenticate by GET
            authentication_url_key (str): key to authenticate by URL. Default is token
            locale (str): Set the locale language. Default is pt-br
            timeout (int): Set the request timeout. Default is 3
            transport_class (obj): Class Request Sync or using AsyncIO
        """
        self.access_token = access_token
        self.accept_language = locale
        self.access_token_type = access_token_type
        self.api_name = api_name
        self.authentication_method = authentication_method
        self.authentication_url_extra_params = authentication_url_extra_params
        self.authentication_url_key = authentication_url_key
        self.base_url = base_url
        self.timeout = timeout
        self._set_transport_class(transport_class=transport_class or RequestsTransport)
        self.kwargs = kwargs

    def _render_response(self, response, endpoint):
        return ResponseFactory(response, endpoint)

    def _set_transport_class(self, transport_class):
        headers = self.get_default_headers()
        self.transport = transport_class(headers=headers)

    def get_default_headers(self):
        """Get the required headers to access the API
        Optional:
            access_token_type (str): Specifies the access_token_type to be used on the request. E.g: Token or Bearer
            accept_language: Specifies the Accept-Language to be receive the response i18n.
        Returns:
             dict: Authorization and Content Type
        """
        return {
            'Authorization': f'{self.access_token_type} {self.access_token}',
            "Accept-Language": self.accept_language
        }

    def make_request(self, method_name, endpoint, **kwargs):
        """Requests data from the API.
        Args:
            endpoint (str): URL for the API endpoint.
            method_name (str): Specifies the method to be used on the request.
        Optional:
            params (dict, optional): Specifies parameters to be sent with the
            request.

        Returns:
            response: Requests response object.
        """

        full_url = urljoin(self.base_url, endpoint)

        content_type = kwargs.pop('content_type', 'application/json')
        if content_type != 'form-data':
            self.transport.session.headers['Content-Type'] = content_type

        json = None
        if content_type == 'application/json':
            json = 'json'

        try:
            response = self.transport.make_request(
                method_name, full_url, timeout=self.timeout, json=json, **kwargs
            )
        except (Timeout, ReadTimeout, RequestsConnectionError) as exc:
            raise APIUnreachableOrOffline(exception_desc=str(exc)) from exc

        is_valid, exception_class = validate_status_code(response.status_code)
        if not is_valid:
            raise exception_class

        return response

    def create(self, endpoint, data={}, files=None, content_type='application/json'):
        """Do a POST without need to pass all arguments to make a request
        Args:
            endpoint (str): URL for the API endpoint.
            data (dict, list of tuples): Data to send in the body of the request.
            files (dict, list of tuples): Files to send in the body of the request.
            content_type (str): Request Content Type. E.g: 'application/json', 'form-data'

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        if content_type == 'application/json' or not files:
            data = json.dumps(data, default=json_converter)
        response = self.make_request('POST', endpoint, data=data, files=files, content_type=content_type)
        return self._render_response(response, endpoint)

    def update(self, endpoint, data={}, files=None, partial=False, content_type='application/json'):
        """Do a update (PUT/PATCH) without need to pass all arguments to make a request
        Args:
            endpoint (str): URL for the API endpoint.
            data (dict, list of tuples): Data to send in the body of the request.
            files (dict, list of tuples): Files to send in the body of the request.
            content_type (str): Request Content Type. E.g: 'application/json', 'form-data'
            partial (bool): To specify whether the update will change everything
                            or just a few attributes. Default is False

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        method = 'PATCH' if partial else 'PUT'
        if content_type == 'application/json':
            data = json.dumps(data, default=json_converter)

        if files:
            response = self.make_request(method, endpoint, data=data, files=files, content_type=content_type)
            return self._render_response(response, endpoint)

        response = self.make_request(method, endpoint, data=data, content_type=content_type)
        return self._render_response(response, endpoint)

    def search(self, endpoint, **kwargs):
        """Do a GET to make a search without need
           to pass all arguments to make a request.
        Args:
            endpoint (str): URL for the API endpoint.
        Kwargs(Optional):
            params (dict): Specifies parameters to be sent with the request.

        Returns:
            dict: Data retrieved for specified endpoint.
        """
        params = kwargs.pop('params', {})
        response = self.make_request('GET', endpoint, params=params)
        return self._render_response(response, endpoint)

    def delete(self, endpoint):
        """Do a DELETE request to remove a resource
        Args:
            endpoint (str): URL for the API endpoint with the ID of the resource.

        Returns:
            dict: Empty
        """
        response = self.make_request('DELETE', endpoint)
        return self._render_response(response, endpoint)


class BaseEndpoint:
    """Class holding endpoint functions.
    """

    def __init__(self, api, endpoint=''):
        self._api = api
        self.endpoint = endpoint
        self.endpoint_name = self._get_endpoint_name()

    def _get_endpoint_name(self):
        words = [key for key in self.endpoint.split('?')[0].split('/') if key]
        response_name = words[-1]
        if '-' in response_name:
            response_name = response_name.replace('-', '_')
        elif response_name.endswith('s'):
            response_name = response_name[:-1]
        return response_name

    def get_endpoint(self, object_id=None, detail=False, *args, **kwargs):
        if detail:
            if not object_id:
                raise APIEndpointMissingArgument(_('Identifier is required.'))
            path = f'{self.endpoint}/{object_id}/'
            if self.endpoint.endswith('/'):
                path = f'{self.endpoint}{object_id}/'
        else:
            path = self.endpoint

        if all([c in path for c in ['{', '}']]):
            try:
                path = path.format(**kwargs)
            except KeyError:
                raise APIEndpointMissingArgument(
                    _('This endpoint contains a Arg in the path and this Arg is required.'))
        return path

    def list(self, *args, **kwargs):
        """Get a list of all Objects of a Resource.

        Args:
            This function takes no arguments.

        Returns:
            dict: Objects of a Resource.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API endpoint.
        """

        endpoint = self.get_endpoint(*args, **kwargs)
        return self._api.search(endpoint, **kwargs)

    def get(self, object_id, *args, **kwargs):
        """Get the full details for a single resource
        Args:
            object_id (str): Object ID.

        Returns:
            dict: Full resource details.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.search(endpoint)

    def create(self, data={}, files=None, content_type='application/json', *args, **kwargs):
        """Update a single Object.

        Args:
            data (dict, list of tuples): Data to send in the body of the request.
            files (dict, list of tuples): Files to send in the body of the request.
            content_type (str): Request Content Type. E.g: 'application/json', 'form-data'

        Returns:
            dict: Data retrieved for specified endpoint.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(*args, **kwargs)
        return self._api.create(endpoint, data=data, files=files, content_type=content_type)

    def update(self, object_id, data={}, files=None, partial=False, content_type='application/json', *args, **kwargs):
        """Update a single Object.

        Args:
            object_id (str): Object ID.
            data (dict, list of tuples): Data to send in the body of the request.
            files (dict, list of tuples): Files to send in the body of the request.
            content_type (str): Request Content Type. E.g: 'application/json', 'form-data'
            partial (bool): To specify whether the update will change everything
                            or just a few attributes. Default is False

        Returns:
            dict: Data retrieved for specified endpoint.

        Raises:
            RequestException: An error thrown by Requests library.
            ValueError: An error thrown by json parser, if JSON decoding fails.
            APIError: An error occurred while requesting the API.

        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.update(endpoint, data=data, files=files, partial=partial, content_type=content_type)

    def delete(self, object_id, *args, **kwargs):
        """Delete a single resource
        Args:
            object_id (str): Object ID.

        Returns:
            dict: Empty

        Raises:
            RequestException: An error thrown by Requests library.
            APIError: An error occurred while requesting the API.
        """
        endpoint = self.get_endpoint(object_id, detail=True, *args, **kwargs)
        return self._api.delete(endpoint)

    def __repr__(self):
        return f'<BaseEndpoint {self.endpoint_name}: {self.endpoint}>'

    def __str__(self):
        return f'{self.endpoint_name} -> {self.endpoint}'



from .exceptions import AuthError, NotFound, ServerError, APIError
from .. import status


def validate_status_code(status_code):
    exception_class = None
    if status_code in (
            status.HTTP_401_UNAUTHORIZED,
            status.HTTP_403_FORBIDDEN):
        exception_class = AuthError(status_code,
                                    {'errorMessage': 'Authentication Error'})
    elif status_code == status.HTTP_404_NOT_FOUND:
        exception_class = NotFound(status_code,
                                   {'errorMessage': 'Endpoint Not Found'})
    elif status_code in range(status.HTTP_405_METHOD_NOT_ALLOWED, status.HTTP_417_EXPECTATION_FAILED + 1):
        exception_class = APIError(status_code)
    elif status.is_server_error(status_code):
        exception_class = ServerError(status_code)

    if exception_class:
        return False, exception_class
    return True, exception_class



import inspect
from math import ceil

from django.core.paginator import EmptyPage, Page, PageNotAnInteger
from django.utils.functional import cached_property
from django.utils.inspect import method_has_no_args
from django.utils.translation import gettext_lazy as _


class ClientAPIPagination:
    def __init__(self, client_method, per_page, **kwargs):
        params =  kwargs.get('extra_params', {})
        params.update({
            'limit': per_page,
            'offset': 0
        })

        self.extra_params = params
        self.client_method = client_method

        self.object_list = client_method(params=self.extra_params).as_obj()
        self.per_page = int(per_page)

    @cached_property
    def count(self):
        """Return the total number of objects, across all pages."""
        c = getattr(self.object_list, 'count', None)
        if callable(c) and not inspect.isbuiltin(c) and method_has_no_args(c):
            return c()
        return self.object_list.count

    @cached_property
    def num_pages(self):
        """Return the total number of pages."""
        if self.count == 0:
            return 0
        hits = max(1, self.count)
        return ceil(hits / self.per_page)

    @property
    def page_range(self):
        """
        Return a 1-based range of pages for iterating through within
        a template for loop.
        """
        return range(1, self.num_pages + 1)

    def validate_number(self, number):
        """Validate the given 1-based page number."""
        try:
            if isinstance(number, float) and not number.is_integer():
                raise ValueError
            number = int(number)
        except (TypeError, ValueError):
            raise PageNotAnInteger(_('That page number is not an integer'))
        if number < 1:
            raise EmptyPage(_('That page number is less than 1'))
        if number > self.num_pages:
            if number == 1:
                pass
            else:
                raise EmptyPage(_('That page contains no results'))
        return number

    def get_page(self, number):
        """
        Return a valid page, even if the page argument isn't a number or isn't
        in range.
        """
        try:
            number = self.validate_number(number)
        except PageNotAnInteger:
            number = 1
        except EmptyPage:
            number = self.num_pages
        return self.page(number)

    def page(self, number):
        """Return a Page object for the given 1-based page number."""
        number = self.validate_number(number)
        offset = (number - 1) * self.per_page
        limit = self.per_page

        params = {
            'limit': limit,
            'offset': offset
        }
        if offset != 0:
            self.extra_params.update(params)
            object_list = self.client_method(params=self.extra_params).as_obj()
        else:
            object_list = self.object_list
        return self._get_page(object_list.results, number, self)

    def _get_page(self, *args, **kwargs):
        """
        Return an instance of a single page.

        This hook can be used by subclasses to use an alternative to the
        standard :cls:`Page` object.
        """
        return Page(*args, **kwargs)



import copy
from django.forms import Select
from django.forms.fields import CallableChoiceIterator, Field
from django.utils.translation import gettext_lazy as _


class AjaxChoiceField(Field):
    widget = Select
    default_error_messages = {
        'invalid_choice': _('Select a valid choice. %(value)s is not one of the available choices.'),
    }

    def __init__(self, *, choices=(), **kwargs):
        super().__init__(**kwargs)
        self.choices = choices

    def __deepcopy__(self, memo):
        result = super().__deepcopy__(memo)
        result._choices = copy.deepcopy(self._choices, memo)
        return result

    def _get_choices(self):
        return self._choices

    def _set_choices(self, value):
        # Setting choices also sets the choices on the widget.
        # choices can be any iterable, but we call list() on it because
        # it will be consumed more than once.
        if callable(value):
            value = CallableChoiceIterator(value)
        else:
            value = list(value)

        self._choices = self.widget.choices = value

    choices = property(_get_choices, _set_choices)

    def to_python(self, value):
        """Return a string."""
        if value in self.empty_values:
            return ''
        return str(value)

    def validate(self, value):
        """Validate that the input is in self.choices."""
        super().validate(value)

    def valid_value(self, value):
        """Check to see if the provided value is a valid choice."""
        text_value = str(value)
        for k, v in self.choices:
            if isinstance(v, (list, tuple)):
                # This is an optgroup, so look inside the group for options
                for k2, v2 in v:
                    if value == k2 or text_value == str(k2):
                        return True
            else:
                if value == k or text_value == str(k):
                    return True
        return False
